import 'dart:async';
import 'dart:convert';

import 'package:collection/collection.dart';
import 'package:flutter/material.dart';
import 'package:neon/blocs.dart';
import 'package:neon/src/models/account.dart';
import 'package:neon/src/models/app_implementation.dart';
import 'package:neon/src/settings/models/storage.dart';
import 'package:neon/src/sync/models/conflicts.dart';
import 'package:neon/src/sync/models/implementation.dart';
import 'package:neon/src/sync/models/mapping.dart';
import 'package:neon/src/utils/sync_mapping_options.dart';
import 'package:rxdart/rxdart.dart';
import 'package:synchronize/synchronize.dart';

abstract interface class SyncBlocEvents {
  /// Adds a new [mapping] that will be synced.
  void addMapping(final SyncMapping<dynamic, dynamic> mapping);

  /// Removes an existing [mapping] that will no longer be synced.
  void removeMapping(final SyncMapping<dynamic, dynamic> mapping);

  /// Explicitly trigger a sync for the [mapping].
  /// [solutions] can be use to apply solutions for conflicts.
  void syncMapping(
    final SyncMapping<dynamic, dynamic> mapping, {
    final Map<String, SyncConflictSolution> solutions = const {},
  });
}

abstract interface class SyncBlocStates {
  /// Map of [SyncMapping]s and their [SyncMappingStatus]es
  BehaviorSubject<Map<SyncMapping<dynamic, dynamic>, SyncMappingStatus>> get mappingStatuses;

  /// Stream of conflicts that have arisen during syncing.
  Stream<SyncConflicts<dynamic, dynamic>> get conflicts;
}

class SyncBloc extends InteractiveBloc implements SyncBlocEvents, SyncBlocStates {
  SyncBloc(
    this._accountsBloc,
    final Iterable<AppImplementation> appImplementations,
  ) {
    _syncImplementations = appImplementations.map((final app) => app.syncImplementation).whereNotNull();
    _timer = TimerBloc().registerTimer(const Duration(minutes: 1), refresh);

    _loadMappings();
    mappingStatuses.value.keys.forEach(_watchMapping);
    unawaited(refresh());
  }

  final AccountsBloc _accountsBloc;
  static const _storage = SingleValueStorage(StorageKeys.sync);
  late final Iterable<SyncImplementation<SyncMapping<dynamic, dynamic>, dynamic, dynamic>> _syncImplementations;
  late final NeonTimer _timer;
  final _conflictsController = StreamController<SyncConflicts<dynamic, dynamic>>();
  final _watchControllers = <String, StreamController<void>>{};
  final _syncMappingOptions = <String, SyncMappingOptions>{};

  @override
  void dispose() {
    _timer.cancel();
    for (final options in _syncMappingOptions.values) {
      options.dispose();
    }
    for (final mapping in mappingStatuses.value.keys) {
      mapping.dispose();
    }
    unawaited(mappingStatuses.close());
    for (final controller in _watchControllers.values) {
      unawaited(controller.close());
    }
    unawaited(_conflictsController.close());

    super.dispose();
  }

  @override
  late final Stream<SyncConflicts<dynamic, dynamic>> conflicts = _conflictsController.stream.asBroadcastStream();

  @override
  final BehaviorSubject<Map<SyncMapping<dynamic, dynamic>, SyncMappingStatus>> mappingStatuses = BehaviorSubject();

  @override
  Future<void> refresh() async {
    for (final mapping in mappingStatuses.value.keys) {
      await _updateMapping(mapping);
    }
  }

  @override
  Future<void> addMapping(final SyncMapping<dynamic, dynamic> mapping) async {
    debugPrint('Adding mapping: $mapping');
    mappingStatuses.add({
      ...mappingStatuses.value,
      mapping: SyncMappingStatus.unknown,
    });
    await _saveMappings();
    // Directly trigger sync after adding the mapping
    await syncMapping(mapping);
    // And start watching for local or remote changes
    _watchMapping(mapping);
  }

  @override
  Future<void> removeMapping(final SyncMapping<dynamic, dynamic> mapping) async {
    debugPrint('Removing mapping: $mapping');
    mappingStatuses.add(Map.fromEntries(mappingStatuses.value.entries.where((final m) => m.key != mapping)));
    mapping.dispose();
    await _saveMappings();
  }

  @override
  Future<void> syncMapping(
    final SyncMapping<dynamic, dynamic> mapping, {
    final Map<String, SyncConflictSolution> solutions = const {},
  }) async {
    debugPrint('Syncing mapping: $mapping');
    mappingStatuses.add({
      ...mappingStatuses.value,
      mapping: SyncMappingStatus.incomplete,
    });

    final account = _accountsBloc.accounts.value.tryFind(mapping.accountId);
    if (account == null) {
      await removeMapping(mapping);
      return;
    }

    try {
      final implementation = _syncImplementations.find(mapping.appId);
      final sources = await implementation.getSources(account, mapping);

      final diff = await computeSyncDiff(
        sources,
        mapping.journal,
        conflictSolutions: solutions,
        keepSkipsAsConflicts: true,
      );
      debugPrint('Journal: ${mapping.journal}');
      debugPrint('Conflicts: ${diff.conflicts}');
      debugPrint('Actions: ${diff.actions}');

      if (diff.conflicts.isNotEmpty && diff.conflicts.whereNot((final conflict) => conflict.skipped).isNotEmpty) {
        _conflictsController.add(
          SyncConflicts<dynamic, dynamic>(
            account,
            implementation,
            mapping,
            diff.conflicts,
          ),
        );
      }

      await executeSyncDiff(sources, mapping.journal, diff);

      mappingStatuses.add({
        ...mappingStatuses.value,
        mapping: diff.conflicts.isEmpty ? SyncMappingStatus.complete : SyncMappingStatus.incomplete,
      });
    } catch (e, s) {
      debugPrint(e.toString());
      debugPrint(s.toString());
      addError(e);
    }

    // Save after syncing even if an error occurred
    await _saveMappings();
  }

  Future<void> _updateMapping(final SyncMapping<dynamic, dynamic> mapping) async {
    final account = _accountsBloc.accounts.value.tryFind(mapping.accountId);
    if (account == null) {
      await removeMapping(mapping);
      return;
    }

    final options = getSyncMappingOptionsFor(mapping);
    if (options.automaticSync.value) {
      await syncMapping(mapping);
    } else {
      try {
        final status = await _getMappingStatus(account, mapping);
        mappingStatuses.add({
          ...mappingStatuses.value,
          mapping: status,
        });
      } catch (e, s) {
        debugPrint(e.toString());
        debugPrint(s.toString());
        addError(e);
      }
    }
  }

  Future<SyncMappingStatus> _getMappingStatus(
    final Account account,
    final SyncMapping<dynamic, dynamic> mapping,
  ) async {
    final implementation = _syncImplementations.find(mapping.appId);
    final sources = await implementation.getSources(account, mapping);
    final diff = await computeSyncDiff(sources, mapping.journal);
    return diff.actions.isEmpty && diff.conflicts.isEmpty ? SyncMappingStatus.complete : SyncMappingStatus.incomplete;
  }

  void _loadMappings() {
    debugPrint('Loading mappings');
    final loadedMappings = <SyncMapping<dynamic, dynamic>>[];

    if (_storage.hasValue()) {
      final serializedMappings = (json.decode(_storage.getString()!) as Map<String, dynamic>)
          .map((final key, final value) => MapEntry(key, (value as List).map((final e) => e as Map<String, dynamic>)));

      for (final mapping in serializedMappings.entries) {
        final syncImplementation = _syncImplementations.tryFind(mapping.key);
        if (syncImplementation == null) {
          continue;
        }

        for (final serializedMapping in mapping.value) {
          loadedMappings.add(syncImplementation.deserializeMapping(serializedMapping));
        }
      }
    }

    mappingStatuses.add({
      for (final mapping in loadedMappings) mapping: SyncMappingStatus.unknown,
    });
  }

  Future<void> _saveMappings() async {
    debugPrint('Saving mappings');
    final serializedMappings = <String, List<Map<String, dynamic>>>{};

    for (final mapping in mappingStatuses.value.keys) {
      final syncImplementation = _syncImplementations.find(mapping.appId);
      serializedMappings[mapping.appId] ??= [];
      serializedMappings[mapping.appId]!.add(syncImplementation.serializeMapping(mapping));
    }

    await _storage.setString(json.encode(serializedMappings));
  }

  void _watchMapping(final SyncMapping<dynamic, dynamic> mapping) {
    final syncImplementation = _syncImplementations.find(mapping.appId);
    if (_watchControllers.containsKey(syncImplementation.getMappingId(mapping))) {
      return;
    }

    // ignore: close_sinks
    final controller = StreamController<void>();
    // Debounce is required to stop bulk operations flooding the sync and potentially creating race conditions.
    controller.stream.debounceTime(const Duration(seconds: 1)).listen((final _) async {
      await _updateMapping(mapping);
    });

    _watchControllers[syncImplementation.getMappingId(mapping)] = controller;

    mapping.watch(() {
      controller.add(null);
    });
  }

  SyncMappingOptions getSyncMappingOptionsFor(final SyncMapping<dynamic, dynamic> mapping) {
    final syncImplementation = _syncImplementations.find(mapping.appId);
    final id = syncImplementation.getGlobalUniqueMappingId(mapping);
    return _syncMappingOptions[id] ??= SyncMappingOptions(
      AppStorage(StorageKeys.sync, id),
    );
  }
}
