import 'package:flutter/material.dart';
import 'package:flutter_material_design_icons/flutter_material_design_icons.dart';
import 'package:meta/meta.dart';
import 'package:neon/l10n/localizations.dart';
import 'package:neon/src/blocs/accounts.dart';
import 'package:neon/src/blocs/sync.dart';
import 'package:neon/src/models/account.dart';
import 'package:neon/src/models/app_implementation.dart';
import 'package:neon/src/pages/sync_mapping_settings.dart';
import 'package:neon/src/settings/widgets/custom_settings_tile.dart';
import 'package:neon/src/settings/widgets/option_settings_tile.dart';
import 'package:neon/src/settings/widgets/settings_category.dart';
import 'package:neon/src/settings/widgets/settings_list.dart';
import 'package:neon/src/theme/dialog.dart';
import 'package:neon/src/utils/confirmation_dialog.dart';
import 'package:neon/src/widgets/account_selection_dialog.dart';
import 'package:neon/src/widgets/sync_status_icon.dart';
import 'package:neon/src/widgets/user_avatar.dart';
import 'package:provider/provider.dart';

@internal
class AppImplementationSettingsPage extends StatelessWidget {
  const AppImplementationSettingsPage({
    required this.appImplementation,
    super.key,
  });

  final AppImplementation appImplementation;

  @override
  Widget build(final BuildContext context) {
    final accountsBloc = Provider.of<AccountsBloc>(context, listen: false);
    final syncBloc = Provider.of<SyncBloc>(context, listen: false);

    final appBar = AppBar(
      title: Text(appImplementation.name(context)),
      actions: [
        IconButton(
          onPressed: () async {
            if (await showConfirmationDialog(
              context,
              NeonLocalizations.of(context).settingsResetForConfirmation(appImplementation.name(context)),
            )) {
              appImplementation.options.reset();
            }
          },
          tooltip: NeonLocalizations.of(context).settingsResetFor(appImplementation.name(context)),
          icon: const Icon(MdiIcons.cogRefresh),
        ),
      ],
    );

    final body = SettingsList(
      categories: [
        for (final category in [...appImplementation.options.categories, null]) ...[
          if (appImplementation.options.options.where((final option) => option.category == category).isNotEmpty) ...[
            SettingsCategory(
              title: Text(
                category != null ? category.name(context) : NeonLocalizations.of(context).optionsCategoryOther,
              ),
              tiles: [
                for (final option
                    in appImplementation.options.options.where((final option) => option.category == category)) ...[
                  OptionSettingsTile(option: option),
                ],
              ],
            ),
          ],
        ],
        if (appImplementation.syncImplementation != null) ...[
          StreamBuilder(
            stream: syncBloc.mappingStatuses,
            builder: (final context, final mappingStatuses) => !mappingStatuses.hasData
                ? const SizedBox.shrink()
                : SettingsCategory(
                    title: Text(NeonLocalizations.of(context).optionsCategorySync),
                    tiles: [
                      for (final mappingStatus in mappingStatuses.requireData.entries
                          .where((final mappingStatus) => mappingStatus.key.appId == appImplementation.id)) ...[
                        CustomSettingsTile(
                          title: Text(appImplementation.syncImplementation!.getMappingDisplayTitle(mappingStatus.key)),
                          subtitle:
                              Text(appImplementation.syncImplementation!.getMappingDisplaySubtitle(mappingStatus.key)),
                          leading: NeonUserAvatar(
                            account: accountsBloc.accounts.value
                                .singleWhere((final account) => account.id == mappingStatus.key.accountId),
                            showStatus: false,
                          ),
                          trailing: IconButton(
                            onPressed: () async {
                              await syncBloc.syncMapping(mappingStatus.key);
                            },
                            tooltip: NeonLocalizations.of(context).syncOptionsSyncNow,
                            iconSize: 30,
                            icon: SyncStatusIcon(
                              status: mappingStatus.value,
                            ),
                          ),
                          onTap: () async {
                            await Navigator.of(context).push(
                              MaterialPageRoute<void>(
                                builder: (final context) => SyncMappingSettingsPage(
                                  mapping: mappingStatus.key,
                                ),
                              ),
                            );
                          },
                        ),
                      ],
                      CustomSettingsTile(
                        title: ElevatedButton.icon(
                          onPressed: () async {
                            final account = await showDialog<Account>(
                              context: context,
                              builder: (final context) => const NeonAccountSelectionDialog(),
                            );
                            if (account == null) {
                              return;
                            }

                            if (!context.mounted) {
                              return;
                            }

                            final mapping = await appImplementation.syncImplementation!.addMapping(context, account);
                            if (mapping == null) {
                              return;
                            }

                            await syncBloc.addMapping(mapping);
                          },
                          icon: const Icon(MdiIcons.cloudSync),
                          label: Text(NeonLocalizations.of(context).syncOptionsAdd),
                        ),
                      ),
                    ],
                  ),
          ),
        ],
      ],
    );

    return Scaffold(
      resizeToAvoidBottomInset: false,
      appBar: appBar,
      body: SafeArea(
        child: Center(
          child: ConstrainedBox(
            constraints: NeonDialogTheme.of(context).constraints,
            child: body,
          ),
        ),
      ),
    );
  }
}
